/*
 * Copyright (C) 2014 Mikel Penagarikano
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package edu.gtts.mediaevalquesst2015;

import java.util.Arrays;
import java.util.LinkedList;

/**
 *
 * @author mpenagar
 */
public class PoolAdjacentViolators {

	/**
	 * PAV: Pool Adjacent Violators algorithm. Non-paramtetric optimization subject to monotonicity.
	 * Fits a vector ghat with nondecreasing components to the data vector y
	 * such that sum((y - ghat).^2) is minimal (Pool-adjacent-violators algorithm).
	 * <p/>
	 * This code is based on the <code>pavx.m</code> from the Bosaris Toolkit
	 * (https://sites.google.com/site/bosaristoolkit/), which was based on the
	 * <code>IsoMeans.m</code> code made available by Lutz Duembgen at:
	 * http://www.imsv.unibe.ch/~duembgen/software
	 * @param y Data vector
	 * @return  A vector containing the transformed values
	 * @see http://www.imsv.unibe.ch/~duembgen/software
	 */
	public static double[] PAV(double[] y) {
		int n = y.length;
		int[] index = new int[n];
		int[] len = new int[n];
		int nw;
		double[] ghat = new double[n];
		int ci = 0;
		index[ci] = 0;
		len[ci] = 1;
		ghat[ci] = y[0];
		for (int j = 1; j < n; j++) {
			index[++ci] = j;
			len[ci] = 1;
			ghat[ci] = y[j];
			while (ci > 0 && ghat[ci - 1] >= ghat[ci]) {
				nw = len[ci - 1] + len[ci];
				ghat[ci - 1] = ghat[ci - 1] + ((double) len[ci] / nw) * (ghat[ci] - ghat[ci - 1]);
				len[--ci] = nw;
			}
		}
		while (n > 0) {
			for (int j = index[ci]; j < n; j++) {
				ghat[j] = ghat[ci];
			}
			n = index[ci--];
		}
		return ghat;
	}

	public static double[] calibrate(double[] score, boolean[] isTarget, Double largeScore, Double prior) {
		int nTar = (largeScore == null) ? 0 : 1;
		if (prior == null) {
			//estimate prior from data
			for (boolean b : isTarget) if (b) nTar++;
			prior = ((double)nTar) / score.length;
		}
		double[] calScore;
		if (largeScore != null) {
			calScore = Arrays.copyOf(score, score.length + 2);
			isTarget = Arrays.copyOf(isTarget, isTarget.length + 2);
			calScore[calScore.length - 2] = -largeScore;
			calScore[calScore.length - 1] = largeScore;
			isTarget[isTarget.length - 2] = true;
			isTarget[isTarget.length - 1] = false;
		} else calScore = score.clone();
		int[] i2i = Util.indexSort(calScore, true);
		double[] prob = new double[calScore.length];
		for (int i = 0; i < calScore.length; i++) {
			prob[i] = isTarget[i2i[i]] ? 1.0 : 0.0;
		}
		prob = PoolAdjacentViolators.PAV(prob);
		double logitPrior = Math.log(prior / (1 - prior));
		for (int i = 0; i < prob.length; i++) {
			calScore[i2i[i]] = Math.log(prob[i] / (1 - prob[i])) - logitPrior;
		}
		return Arrays.copyOf(calScore, score.length);
	}

	private static void fastPAV(double[] x) {
		int k;
		int m;
		int u;
		boolean flag = false;
		int N = x.length;
		int[] F = new int[N];
		int[] B = new int[N];
		for (k = 0; k < N; k++) {
			F[k] = k + 1;
			B[k] = k - 1;
		}
		LinkedList<Integer> Q1 = new LinkedList<Integer>();
		LinkedList<Integer> Q2 = new LinkedList<Integer>();
		LinkedList<Integer> q = null;
		for (k = 0; k < N; k++) {
			Q1.add(k);
		}
		while (!Q1.isEmpty()) {
			m = 0;
			while (!Q1.isEmpty()) {
				k = Q1.removeFirst();
				if (m <= k) {
					flag = false;
					System.out.println("k:" + k + " F[k]:" + F[k]);
					while (F[k] != k && x[k] > x[F[k]]) {
						System.out.println("Entra");
						u = F[F[k]];
						F[k] = u;
						if (u < N) {
							B[u] = k;
						}
						m = u;
						flag = true;
						System.out.println(" --> u:" + u + " k:" + k + " F[k]:" + F[k]);
					}
				}
				if (flag && B[k] >= 0) {
					Q2.add(B[k]);
				}
			}
			q = Q1;
			Q1 = Q2;
			Q2 = q;
		}
	}
	
}
