/*
 * Copyright (C) 2014 Mikel Penagarikano
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package edu.gtts.mediaevalquesst2015;

import java.util.Arrays;


/**
 * Computes Pooled Detection Error Tradeoff values (Pmiss & Pfa)
 * @author mpenagar
 */
public class PooledDET {
	
	int trials;
	double[] pMiss = null;
	double[] pFA = null;
	double[] threshold = null;
	
	private PooledDET(int trials) {
		this.trials = trials;
		pMiss = new double[trials + 1];
		pFA = new double[trials + 1];
		threshold = new double[trials + 1];
		// below minimum score, pMiss=1 & pFA=0
		pMiss[0] = 0.0;
		pFA[0] = 1.0;
		threshold[0] = Double.NEGATIVE_INFINITY;
	}
	
	private void shrink(int newLength){
		if (newLength != pMiss.length) {
			pMiss = Arrays.copyOf(pMiss, newLength);
			pFA = Arrays.copyOf(pFA, newLength);
			threshold = Arrays.copyOf(threshold, newLength);
		}
	}

	/**
	 * Computes Pooled Detection Error Tradeoff values (Pmiss & Pfa)
	 * @param score
	 * @param modelIndex
	 * @param isTarget
	 * @return 
	 */
	public static PooledDET fromScores(double[] score,int[] modelIndex,boolean[] isTarget) {

		double[] pmissIncrement = null;
		double[] pfaDecrement = null;

		{
			int[][] counts = modelCounts(modelIndex,isTarget);
			int[] tarCounts = counts[0];
			int[] nonCounts = counts[1];
			int nModel = tarCounts.length;
			pmissIncrement = new double[nModel];
			pfaDecrement = new double[nModel];
			for (int i = 0; i < nModel; i++) {
				pmissIncrement[i] = (1.0 / tarCounts[i]) / nModel;
				pfaDecrement[i] = (1.0 / nonCounts[i]) / nModel;
			}
		}
		
		PooledDET det = new PooledDET(score.length);
		System.arraycopy(score, 0, det.threshold, 1, score.length);
		int[] sortedIndex = Util.indexSort(score,true);

		int j = 0;
		for (int i = 0; i < score.length; i++){
			int scoreIndex = sortedIndex[i];
			int m = modelIndex[scoreIndex];
			// avoid saving det-points with same scores 
			if (score[scoreIndex] == det.threshold[j]) {
				if (isTarget[scoreIndex]) det.pMiss[j] += pmissIncrement[m];
				else det.pFA[j] -= pfaDecrement[m];
			} else {
				j++;
				if (isTarget[scoreIndex]) {
					det.pMiss[j] = det.pMiss[j-1] + pmissIncrement[m];
					det.pFA[j] = det.pFA[j-1];
				} else {
					det.pMiss[j] = det.pMiss[j-1];
					det.pFA[j] = det.pFA[j-1] - pfaDecrement[m];
				}
				det.threshold[j] = score[scoreIndex];
			}
		}
		det.shrink(j+1);
		return det;
	}
	
	private static int[][] modelCounts(int[] modelIndex,boolean[] isTarget){
		int max = 0;
		int[] tar = new int[100];
		int[] non = new int[100];
		Arrays.fill(tar,0);
		Arrays.fill(non,0);
		for (int i = 0; i < modelIndex.length; i++) {
			int m = modelIndex[i];
			if (m > max) max = m;
			if (m >= tar.length) {
				tar = Arrays.copyOf(tar,(m * 4) / 3);
				non = Arrays.copyOf(non,tar.length);
			}
			if (isTarget[i]) tar[m]++;
			else non[m]++;
		}
		int[][] tarnon = {Arrays.copyOf(tar,max+1),Arrays.copyOf(non,max+1)};
		return tarnon;
	}
	
	
	public static void main(String[] args) {
		double[] s =  {2,   1,    7,   -1,  3,   2,    30,  4,    3,   2};
		int[] m =     {0,   0,    0,   0,   1,   1,    2,   2,    3,   3};
		boolean[] t = {true,false,true,true,true,false,true,false,true,false};

		PooledDET det = fromScores(s,m,t);
		
		System.out.println(Arrays.toString(det.pMiss));
		System.out.println(Arrays.toString(det.pFA));
		
	}
}
