/*
 * Copyright (C) 2014 Mikel Penagarikano
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package edu.gtts.mediaevalquesst2015;

import johnacklam.StatUtil;
import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.util.Arrays;

/**
 *
 * @author Mikel Penagarikano
 */
public class TermWeightedValue {
	private int maxIndex;
	public PooledDET det;
	public double beta;
	public double[] value;
	public double actualValue;

	/**
	 * Estimates the actual Term-Weighted Value (ATWV) based on a set of decisions.
	 * ATWV represents the system’s ability to predict the optimal operating point
	 * given the TWV scoring metric
	 * @param modelIndex Class (label) indexes of each trial 
	 * @param isTarget Ground truth of each trial
	 * @param decision Actual decision of each trial
	 * @param beta The Beta weighing parameter
	 * @return the actual Term Weighted Value based on decisions.
	 */
	public static double getActual(int[] modelIndex,boolean[] isTarget,boolean[] decision, double beta){
		int max = 0;
		int[] tar = new int[100];
		int[] non = new int[100];
		int[] miss = new int[100];
		int[] fa = new int[100];
		for (int i = 0; i < modelIndex.length; i++) {
			int m = modelIndex[i];
			if (m > max) max = m;
			if (m >= tar.length) {
				tar = Arrays.copyOf(tar,(m * 4) / 3);
				non = Arrays.copyOf(non,tar.length);
				miss = Arrays.copyOf(miss,tar.length);
				fa = Arrays.copyOf(fa,tar.length);
			}
			if (isTarget[i]) {
				tar[m]++;
				if (!decision[i]) miss[m]++;
			} else {
				non[m]++;
				if (decision[i]) fa[m]++;
			}
		}
		double sum = 0.0;
		for (int m = 0; m <= max; m++)
			sum += (((double)miss[m])/tar[m] + beta * ((double)fa[m])/non[m]);
		return 1.0 - sum / (max+1);
	}
	
	public TermWeightedValue(double[] score,int[] modelIndex,boolean[] isTarget,boolean[] decision, double beta) {
		this.actualValue = getActual(modelIndex,isTarget,decision,beta);
		this.det = PooledDET.fromScores(score,modelIndex,isTarget);
		this.beta = beta;
		value = new double[det.pMiss.length];
		double[] pMiss = det.pMiss;
		double[] pFA = det.pFA;
		double max = 1.0 - beta;
		maxIndex = 0;
		
		for (int i = 0; i < value.length; i++) {
			value[i] = 1.0 - (pMiss[i] + beta * pFA[i]);
			if (value[i] > max) {
				max = value[i];
				maxIndex = i;
			}
		}
	}	

	public double maxValue(){
		return value[maxIndex];
	}
	public double maxValueThreshold(){
		return det.threshold[maxIndex];
	}
	public double maxValuePmiss(){
		return det.pMiss[maxIndex];
	}
	public double maxValuePfa(){
		return det.pFA[maxIndex];
	}

	/**
	 * Round probabilities to 5 decimals
	 * @param p the probability
	 * @return the rounded probability
	 */
	private double round(double p){
		return (double)Math.round(p * 100000.0) / 100000.0;
	}

	public void writePlotFiles(String systemName) {
		PrintWriter out = null;
		String fileName = null;
		
		// plot.TWV.data
		try {
			fileName = "plot.TWV.data";
			out = new PrintWriter(fileName);
			out.println(
				"# DET Graph made by MediaEvalQUESST2014\n" +
				"# PooledTotalTrials = " + det.trials + "\n" +
				"# DET Type: Term Wtd.\n" +
				"# Abbreviations: ssd() is the sample Standard Deviation of a Variable\n" +
				"#                ppndf() is the normal deviant of a probability. ppndf(.5)=0\n" +
				"#                -2SE(v) is v - 2(StandardError(v)) = v - 2 * (sampleStandardDev / sqrt(n))\n" +
				"# score ppndf(Pmiss) ppndf(Pfa) Pmiss Pfa Value ppndf(-2SE(Pmiss)) ppndf(-2SE(Pfa)) ppndf(+2SE(Pmiss)) ppndf(+2SE(Pfa)) SE(Value)"
			);
			for (int i = 0; i < det.pMiss.length; i++)
				out.println(
					(float)det.threshold[i] +  " " + 
					(float)invCDF(det.pMiss[i]) + " " + (float)invCDF(det.pFA[i]) + " " + 
					(float)round(det.pMiss[i]) + " " + (float)round(det.pFA[i]) + " " + 
					(float)value[i] +
					" - - - - -"
				) ;
			out.close();
		} catch (FileNotFoundException ex) {
			System.err.println("Could not write to file: " + fileName + "\n\nEXITING...");
			System.exit(1);
		}
		
		// plot.maxTWV.data
		try {
			fileName = "plot.maxTWV.data";
			out = new PrintWriter(fileName);
			out.println(
				"# Points for DET Graph made by MediaEvalQUESST2014\n" +
				"# DET Type: Term Wtd.\n" +
				"# MaxValueDetectionScore MaxValueValue Pmiss Pfa ppndf(Pmiss) ppndf(Pfa) ActualValue"
			);
			out.println(
				(float)maxValueThreshold() + " " + (float)maxValue() + " " +
				(float)maxValuePmiss() + " " + (float)maxValuePfa() + " " +
				(float)invCDF(maxValuePmiss()) + " " + (float)invCDF(maxValuePfa())
			);
			out.close();
		} catch (FileNotFoundException ex) {
			System.err.println("Could not write to file: " + fileName + "\n\nEXITING...");
			System.exit(1);
		}
		
		// plot.DET.plt
		try {
			fileName = "plot.DET.plt";
			out = new PrintWriter(fileName);
			out.println(
				"## GNUPLOT command file\n" +
				"set terminal postscript enhanced color\n" +
				"set style data lines\n" +
				"set noxzeroaxis\n" +
				"set noyzeroaxis\n" +
				"set key top spacing 1.5\n" +
				"set size ratio 0.821894871074622\n" +
				"set noxtics\n" +
				"set noytics\n" +
				"set title 'Term Wtd. DET Curve (" + systemName + ")'\n" +
				"set xlabel 'False Alarm probability (in %)'\n" +
				"set ylabel 'Miss probability (in %)'\n" +
				"set grid\n" +
				"set pointsize 3\n" +
				"set ytics (\\\n" +
				"    '5' -1.6449, '10' -1.2816, '20' -0.8416, '40' -0.2533, '60' 0.2533, \\\n" +
				"    '80' 0.8416, '90' 1.2816, '95' 1.6449, '98' 2.0537)\n" +
				"set xtics (\\\n" +
				"    '.0001' -4.7534, '.001' -4.2649, '.004' -3.9444, '.01' -3.7190, '.02' -3.5401, \\\n" +
				"    '.05' -3.2905, '.1' -3.0902, '.2' -2.8782, '.5' -2.5758, '1' -2.3263, \\\n" +
				"    '2' -2.0537, '5' -1.6449, '10' -1.2816, '20' -0.8416, '40' -0.2533)\n" +
				"plot [-4.75343910607888:-0.253347103317183] [-1.64485362793551:2.05374890849825] \\\n" +
				"   -x title 'Random Performance' with lines lc 1,\\\n" +
				"    'plot.TWV.data' using 3:2 title 'MaxTWV " + String.format("%.4g",maxValue()) + " @ " + String.format("%.4g",maxValueThreshold()) + " (pFA=" + String.format("%.4g",100*maxValuePfa())+ ", pMiss=" + String.format("%.4g",100*maxValuePmiss())+ ")' with lines lc 2,\\\n" +
				"    'plot.maxTWV.data' using 6:5 notitle w p lc 2 pt 6 ps 2"
			);
			out.close();
		} catch (FileNotFoundException ex) {
			System.err.println("Could not write to file: " + fileName + "\n\nEXITING...");
			System.exit(1);
		}
		
		// plot.TWV.plt
		try {
			fileName = "plot.TWV.plt";
			out = new PrintWriter(fileName);
			out.println(
				"## GNUPLOT command file\n" +
				"set terminal postscript enhanced color\n" +
				"set style data lines\n" +
				"set key top spacing 1.5\n" +
				"set title 'Term Wtd. Threshold Plot (" + systemName + ")'\n" +
				"set xlabel 'Decision Threshold'\n" +
				"set ylabel 'Term Weighted Value  ({/Symbol b}=" + String.format("%.5g",beta) + ")'\n" +
				"set grid\n" +
				"plot [" + det.threshold[1] + ":" + det.threshold[det.threshold.length-1] + "] [0:1] \\\n" +
				"  'plot.TWV.data' using 1:4 title 'P(Miss)', \\\n" +
				"  'plot.TWV.data' using 1:5 title 'P(FA)', \\\n" +
				"  'plot.TWV.data' using 1:6 title 'Value', \\\n" +
				"  'plot.maxTWV.data' using 1:2 title 'MaxTWV " + String.format("%.4g",maxValue()) + " @ " + String.format("%.4g",maxValueThreshold()) + "' w p pt 6 ps 2"
			);
			out.close();
		} catch (FileNotFoundException ex) {
			System.err.println("Could not write to file: " + fileName + "\n\nEXITING...");
			System.exit(1);
		}
	}
	
	private static final double LOW_PROBABILITY = 1e-15;
	private static final double HIGH_PROBABILITY = 1 - LOW_PROBABILITY;
	public static double invCDF(double prob) {
		return	prob < LOW_PROBABILITY ? StatUtil.getInvCDF(LOW_PROBABILITY,false) :
				prob > HIGH_PROBABILITY ? StatUtil.getInvCDF(HIGH_PROBABILITY,false) :
				StatUtil.getInvCDF(prob,false);
	}
	
	public static void main(String[] args) {
		
		System.out.println(String.format("%.4f",0.0021321));
		System.out.println(String.format("%.4f",23.423123));
		System.out.println(String.format("%.4g",0.0021321));
		System.out.println(String.format("%.4g",23.423123));
	}
}
